"""
========================================================================
AstHelper.py
========================================================================
This file collects PyMTL ast visitors.

Author : Shunning Jiang
Date   : Jan 17, 2018
"""

import ast

from .errors import InvalidFFAssignError


class DetectVarNames( ast.NodeVisitor ):

  def __init__( self, upblk, obj, is_update_ff ):
    self.upblk = upblk
    self.obj = obj
    self.is_update_ff = is_update_ff
    self.globals = upblk.__globals__
    self.closure = { *upblk.__code__.co_freevars }

  # Helper function to get the full name containing "s"

  def _get_full_name( self, node ):

    # We store the name/index linearly, and store the corresponding ast
    # nodes linearly for annotation purpose
    obj_name = []
    nodelist = []

    # First strip off all slices
    # s.x[1][2].y[i][3]
    slices = []
    while isinstance( node, ast.Subscript ) and isinstance( node.slice, ast.Slice ):
      lower = node.slice.lower
      upper = node.slice.upper
      # If the slice looks like a[i:i+1] where i is variable, I assume it
      # would access the whole variable

      # Shunning: since the closure/global variables can vary across
      # different instances of the same class, we need to cache the name.

      low = up = None

      if isinstance( lower, ast.Num ):
        low = node.slice.lower.n
      elif isinstance( lower, ast.Name ):
        x = lower.id
        if   x in self.globals:
          low = (False, x)
        elif x in self.closure:
          low = (True, x)

      if isinstance( upper, ast.Num ):
        up = node.slice.upper.n
      elif isinstance( upper, ast.Name ):
        x = upper.id
        if   x in self.globals:
          up = (False, x)
        elif x in self.closure:
          up = (True, x)

      if low is not None and up is not None:
        slices.append( slice(low, up) )
      # FIXME
      # else:

      nodelist.append( node )
      node = node.value

    # s.x[1][2].y[i]
    while True:
      num = []
      while isinstance( node, ast.Subscript ) and \
            isinstance( node.slice, ast.Index ):
        v = node.slice.value
        n = "*"

        if   isinstance( v, ast.Num ):
          n = v.n
        elif isinstance( v, ast.Name ):
          x = v.id
          if   x in self.globals: # Only support global const indexing for now
            n = (False, x)
          elif x in self.closure:
            n = (True, x)
        elif isinstance( v, ast.Attribute ): # s.sel, may be constant
          self.visit( v )
        elif isinstance( v, ast.Call ): # int(x)
          for x in v.args:
            self.visit(x)

        num.append(n)

        nodelist.append( node )
        node = node.value

      if   isinstance( node, ast.Attribute ):
        obj_name.append( (node.attr, num[::-1]) )
      elif isinstance( node, ast.Name ):
        obj_name.append( (node.id, num[::-1]) )
      elif isinstance( node, ast.Call ): # a.b().c()
        # FIXME?
        return None, None
      else:
        assert isinstance( node, ast.Str ) # filter out line_trace
        return None, None

      nodelist.append( node )

      if not hasattr( node, "value" ):
        break
      node = node.value

    if slices:
      assert len(slices) == 1, "Multiple slices at the end of s.%s in update block %s" % \
        ( ".".join( [ obj_name[i][0] + "".join(["[%s]" % x for x in obj_name[i][1]]) for i in range(len(obj_name)) ] ) \
        +  "[%d:%d]" % (x[0], x[1]), self.upblk.__name__ )

      obj_name[0][1].append( slices[0] )

    obj_name = obj_name[::-1]
    nodelist = nodelist[::-1]
    return obj_name, nodelist

class DetectReadsWritesCalls( DetectVarNames ):

  def enter( self, node, read, write, calls ):

    self.read = read
    self.write = write
    self.calls = calls
    self.visit( node )

  def visit_Assign( self, node ):

    if self.is_update_ff:
      assert len( node.targets ) == 1
      target = node.targets[0]

      if isinstance( target, (ast.Attribute, ast.Subscript) ):
        while isinstance( target, (ast.Attribute, ast.Subscript) ):
          target = target.value
        assert isinstance( target, ast.Name ), "Please call pymtl3 developers"

        if target.id == "s":
          raise InvalidFFAssignError( self.obj, self.upblk, node.lineno,
                "has a wrong assign operator. Change it to <<=." )

    for x in node.targets:
      self.visit( x )
    self.visit( node.value )

  def visit_AugAssign( self, node ):
    if self.is_update_ff:
      if isinstance( node.target, (ast.Attribute, ast.Subscript) ):
        if not isinstance( node.op, ast.LShift ):
          raise InvalidFFAssignError( self.obj, self.upblk, node.lineno,
                "has a wrong assign operator. Change it to <<=." )

    self.visit( node.target )
    self.visit( node.value  )

  def visit_Attribute( self, node ): # s.a.b
    obj_name, nodelist = self._get_full_name( node )
    if not obj_name:  return

    pair = (obj_name, nodelist)

    if   isinstance( node.ctx, ast.Load ):
      self.read.append( pair )
    elif isinstance( node.ctx, ast.Store ):
      self.write.append( pair )
    else:
      raise TypeError( f"Wrong ast node context {type( node.ctx )}" )

  def visit_Subscript( self, node ): # s.a.b[0:3] or s.a.b[0]
    obj_name, nodelist = self._get_full_name( node )
    if not obj_name:  return

    pair = (obj_name, nodelist)

    if   isinstance( node.ctx, ast.Load ):
      self.read.append( pair )
    elif isinstance( node.ctx, ast.Store ):
      self.write.append( pair )
    else:
      raise TypeError( f"Wrong ast node context {type( node.ctx )}" )

    self.visit( node.slice )

  def visit_Call( self, node ):
    obj_name, nodelist = self._get_full_name( node.func )
    if not obj_name:  return

    self.calls.append( (obj_name, nodelist) )

    for x in node.args:
      self.visit( x )

class DetectMethodCalls( DetectVarNames ):

  def enter( self, node, methods ):
    self.methods = methods
    self.visit( node )

  def visit_Call( self, node ):
    obj_name = self.get_full_name( node.func )
    if not obj_name: return # to check node.func.id

    pair = (obj_name, node)

    self.methods.append( pair )

    for x in node.args:
      self.visit( x )

def extract_reads_writes_calls( hostobj, f, tree, is_update_ff, read, write, calls ):

  # Traverse the ast to extract variable writes and reads
  # First check and remove @s.update and empty arguments
  assert isinstance(tree, ast.Module)
  tree = tree.body[0]
  assert isinstance(tree, ast.FunctionDef)

  visitor = DetectReadsWritesCalls( f, hostobj, is_update_ff )
  for stmt in tree.body:
    visitor.enter( stmt, read, write, calls )

def get_method_calls( tree, upblk, methods ):
  DetectMethodCalls( upblk, hostobj, is_update_ff ).enter( tree, methods )



#=========================================================================
# BehavioralRTLIRTypeCheckL5Pass.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 30, 2019
"""Provide L5 behavioral RTLIR type check pass."""

from collections import OrderedDict

from pymtl3.passes.BasePass import BasePass, PassMetadata
from pymtl3.passes.rtlir.errors import PyMTLTypeError
from pymtl3.passes.rtlir.rtype import RTLIRType as rt

from .BehavioralRTLIRTypeCheckL4Pass import BehavioralRTLIRTypeCheckVisitorL4


class BehavioralRTLIRTypeCheckL5Pass( BasePass ):
  def __call__( s, m ):
    """Perform type checking on all RTLIR in rtlir_upblks."""
    if not hasattr( m, '_pass_behavioral_rtlir_type_check' ):
      m._pass_behavioral_rtlir_type_check = PassMetadata()
    m._pass_behavioral_rtlir_type_check.rtlir_freevars = OrderedDict()
    m._pass_behavioral_rtlir_type_check.rtlir_tmpvars = OrderedDict()
    m._pass_behavioral_rtlir_type_check.rtlir_accessed = set()

    visitor = BehavioralRTLIRTypeCheckVisitorL5( m,
      m._pass_behavioral_rtlir_type_check.rtlir_freevars,
      m._pass_behavioral_rtlir_type_check.rtlir_accessed,
      m._pass_behavioral_rtlir_type_check.rtlir_tmpvars
    )

    for blk in m.get_update_block_order():
      visitor.enter( blk, m._pass_behavioral_rtlir_gen.rtlir_upblks[ blk ] )

class BehavioralRTLIRTypeCheckVisitorL5( BehavioralRTLIRTypeCheckVisitorL4 ):
  def __init__( s, component, freevars, accessed, tmpvars ):
    super(). \
        __init__( component, freevars, accessed, tmpvars )

  def visit_Index( s, node ):
    """Type check the index node.

    Only static constant expressions can be the index of component arrays
    """
    if isinstance( node.value.Type, rt.Array ) and \
       isinstance( node.value.Type.get_sub_type(), rt.Component ):
      node.Type = node.value.Type.get_sub_type()

    else:
      super().visit_Index( node )

  def visit_Attribute( s, node ):
    """Type check an attribute.

    Since only the ports of a subcomponent can be accessed, no explicit
    cross-hierarchy access detection is needed.
    """
    # Attributes of subcomponent can only access ports
    if isinstance( node.value.Type, rt.Component ) and \
       node.value.Type.get_name() != s.component.__class__.__name__:
      if not node.value.Type.has_property( node.attr ):
        raise PyMTLTypeError( s.blk, node.ast,
          f'rt.Component {node.value.Type.get_name()} does not have attribute {node.attr}!' )
      prop = node.value.Type.get_property( node.attr )
      if not rt._is_of_type( prop, ( rt.Port, rt.InterfaceView ) ):
        raise PyMTLTypeError( s.blk, node.ast,
          f'{node.attr} is not a port of {node.value.Type.get_name()} subcomponent!' )
      node.Type = prop
    else:
      super().visit_Attribute( node )



from .bits_import import *
from .bits_import import _bitwidths
from .bitstructs import bitstruct, is_bitstruct_class, is_bitstruct_inst, mk_bitstruct
from .helpers import (
    clog2,
    concat,
    get_bitstruct_inst_all_classes,
    get_nbits,
    reduce_and,
    reduce_or,
    reduce_xor,
    sext,
    to_bits,
    zext,
)



"""
#=========================================================================
# MemIfcs
#=========================================================================
#
# Author: Shunning Jiang
# Date  : May 18, 2019
"""
from greenlet import greenlet

from pymtl3 import *
from pymtl3.stdlib.connects import connect_pairs

from .MemMsg import MemMsgType, mk_mem_msg
from .SendRecvIfc import RecvCL2SendRTL, RecvIfcRTL, RecvRTL2SendCL, SendIfcRTL


class MemMasterIfcFL( Interface ):
  def construct( s ):
    s.read  = CallerIfcFL()
    s.write = CallerIfcFL()
    s.amo   = CallerIfcFL()

  def __str__( s ):
    return f"r{s.read}|w{s.write}|a{s.amo}"

  def connect( s, other, parent ):
    if isinstance( other, MemMinionIfcCL ):
      m = MemIfcFL2CLAdapter( other.ReqType, other.RespType )

      if hasattr( parent, "MemIfcFL2CL_count" ):
        count = parent.MemIfcFL2CL_count
        setattr( parent, "MemIfcFL2CL_" + str( count ), m )
      else:
        parent.MemIfcFL2CL_count = 0
        parent.MemIfcFL2CL_0 = m

      connect_pairs(
        s,       m.left,
        m.right, other,
      )
      parent.MemIfcFL2CL_count += 1
      return True

    elif isinstance( other, MemMinionIfcRTL ):
      m = MemIfcFL2RTLAdapter( other.ReqType, other.RespType )

      if hasattr( parent, "MemIfcFL2RTL_count" ):
        count = parent.MemIfcFL2RTL_count
        setattr( parent, "MemIfcFL2RTL_" + str( count ), m )
      else:
        parent.MemIfcFL2RTL_count = 0
        parent.MemIfcFL2RTL_0 = m

      connect_pairs(
        s,       m.left,
        m.right, other,
      )
      parent.MemIfcFL2RTL_count += 1
      return True

    return False

class MemMinionIfcFL( Interface ):
  def construct( s, read=None, write=None, amo=None ):
    s.read  = CalleeIfcFL( method=read )
    s.write = CalleeIfcFL( method=write )
    s.amo   = CalleeIfcFL( method=amo )

  def __str__( s ):
    return f"r{s.read}|w{s.write}|a{s.amo}"

  def connect( s, other, parent ):
    if isinstance( other, MemMasterIfcCL ):
      m = MemIfcCL2FLAdapter( other.ReqType, other.RespType )

      if hasattr( parent, "MemIfcCL2FL_count" ):
        count = parent.MemIfcCL2FL_count
        setattr( parent, "MemIfcCL2FL_" + str( count ), m )
      else:
        parent.MemIfcCL2FL_count = 0
        parent.MemIfcCL2FL_0 = m

      connect_pairs(
        other,   m.left,
        m.right, other,
      )
      parent.MemIfcCL2FL_count += 1
      return True

    elif isinstance( other, MemMasterIfcRTL ):
      m = MemIfcRTL2FLAdapter( other.ReqType, other.RespType )

      if hasattr( parent, "MemIfcRTL2FL_count" ):
        count = parent.MemIfcRTL2FL_count
        setattr( parent, "MemIfcRTL2FL_" + str( count ), m )
      else:
        parent.MemIfcRTL2FL_count = 0
        parent.MemIfcRTL2FL_0 = m

      connect_pairs(
        other,   m.left,
        m.right, s,
      )
      parent.MemIfcRTL2FL_count += 1
      return True

    return False

class MemMasterIfcCL( Interface ):
  def construct( s, ReqType, RespType, resp=None, resp_rdy=None ):
    s.ReqType  = ReqType
    s.RespType = RespType
    s.req  = CallerIfcCL( Type=ReqType )
    s.resp = CalleeIfcCL( Type=RespType, method=resp, rdy=resp_rdy )

  def line_trace( s ):
    return "{} > {}".format( s.req, s.resp )

  def connect( s, other, parent ):
    if isinstance( other, MemMinionIfcCL ):
      assert s.ReqType is other.ReqType and s.RespType is other.RespType
    return False

class MemMinionIfcCL( Interface ):
  def construct( s, ReqType, RespType, req=None, req_rdy=None ):
    s.ReqType  = ReqType
    s.RespType = RespType
    s.req  = CalleeIfcCL( Type=ReqType, method=req, rdy=req_rdy )
    s.resp = CallerIfcCL( Type=RespType )

  def line_trace( s ):
    return "{} > {}".format( s.req, s.resp )

class MemMasterIfcRTL( Interface ):

  def construct( s, ReqType, RespType ):
    s.ReqType  = ReqType
    s.RespType = RespType
    s.req  = SendIfcRTL( ReqType  )
    s.resp = RecvIfcRTL( RespType )

  def __str__( s ):
    return "{},{}".format( s.req, s.resp )

class MemMinionIfcRTL( Interface ):
  def construct( s, ReqType, RespType ):
    s.ReqType  = ReqType
    s.RespType = RespType
    s.req  = RecvIfcRTL( ReqType  )
    s.resp = SendIfcRTL( RespType )

  def __str__( s ):
    return "{},{}".format( s.req, s.resp )

class MemIfcCL2FLAdapter( Component ):

  def recv_rdy( s ):
    return s.entry is None

  def recv( s, msg ):
    assert s.entry is None
    s.entry = msg

  def construct( s, ReqType, RespType ):
    s.left  = MemMinionIfcCL( ReqType, RespType, s.recv, s.recv_rdy )
    s.right = MemMasterIfcFL()
    s.entry = None

    @s.update
    def up_memifc_cl_fl_blk():

      if s.entry is not None and s.left.resp.rdy():

        # Dequeue memory request message

        req     = s.entry
        s.entry = None

        len_ = int(req.len)
        if not len_: len_ = ReqType.data_nbits >> 3

        if   req.type_ == MemMsgType.READ:
          resp = RespType( req.type_, req.opaque, 0, req.len,
                             s.right.read( req.addr, len_ ) )

        elif req.type_ == MemMsgType.WRITE:
          s.right.write( req.addr, len_, req.data )
          # FIXME do we really set len=0 in response when doing subword wr?
          # resp = RespTypees( req.type_, req.opaque, 0, req.len, 0 )
          resp = RespType( req.type_, req.opaque, 0, 0, 0 )

        else: # AMOS
          resp = RespType( req.type_, req.opaque, 0, req.len,
             s.right.amo( req.type_, req.addr, len_, req.data ) )

        # Make line trace look better since s.right might get blocked
        assert s.left.resp.rdy()
        s.left.resp( resp )

    s.add_constraints(
      M(s.left.req) < U(up_memifc_cl_fl_blk), # bypass behavior
    )

class MemIfcFL2CLAdapter( Component ):

  def read( s, addr, nbytes ):

    # TODO refactor this greenlet stuff into some utility API
    while not s.right.req.rdy():
      greenlet.getcurrent().parent.switch(0)

    s.right.req( s.ReqType( MemMsgType.READ, 0, addr, nbytes ) )

    while s.entry is None:
      greenlet.getcurrent().parent.switch(0)

    ret = s.entry.data
    s.entry = None
    return ret

  def write( s, addr, nbytes, data ):

    while not s.right.req.rdy():
      greenlet.getcurrent().parent.switch(0)

    s.right.req( s.ReqType( MemMsgType.WRITE, 0, addr, nbytes, data ) )

    while s.entry is None:
      greenlet.getcurrent().parent.switch(0)

    s.entry = None

  def amo( s, amo, addr, nbytes, data ):

    while not s.right.req.rdy():
      greenlet.getcurrent().parent.switch(0)

    s.right.req( s.ReqType( amo, 0, addr, nbytes ) )

    while s.entry is None:
      greenlet.getcurrent().parent.switch(0)

    ret = s.entry.data
    s.entry = None
    return ret

  def recv_rdy( s ):
    return s.entry is None

  def recv( s, msg ):
    assert s.entry is None
    s.entry = msg

  def construct( s, ReqType, RespType ):
    s.entry = None # store response

    s.ReqType  = ReqType
    s.RespType = RespType

    s.left  = MemMinionIfcFL( read=s.read, write=s.write, amo=s.amo )
    s.right = MemMasterIfcCL( ReqType, RespType, s.recv, s.recv_rdy )

    s.add_constraints(
      M(s.left.read)  == M(s.right.req),
      M(s.left.write) == M(s.right.req),
      M(s.left.amo)   == M(s.right.req),
      M(s.left.read)  >  M(s.right.resp), # Comb behavior
      M(s.left.write) >  M(s.right.resp), # Comb behavior
      M(s.left.amo)   >  M(s.right.resp), # Comb behavior
    )

#-------------------------------------------------------------------------
# RTL/FL adapters
#-------------------------------------------------------------------------

class MemIfcRTL2FLAdapter( Component ):

  def construct( s, ReqType, RespType ):
    s.left  = MemMinionIfcRTL( ReqType, RespType )
    s.right = MemMasterIfcFL()

    @s.update
    def up_memifc_rtl_fl_blk():

      if s.left.req.en and s.left.resp.rdy:

        if s.left.req.msg.type_ == MemMsgType.READ:
          resp = RespType( s.left.req.msg.type_, s.right.read( s.left.req.msg.addr ) )

        elif s.left.req.msg.type_ == MemMsgType.WRITE:
          s.right.write( s.left.req.msg.addr, s.left.req.msg.data )
          resp = RespType( s.left.req.msg.type_, 0 )

        else: # AMOs
          resp = RespType( req.type_, req.opaque, 0, req.len,
             s.right.amo( req.type_, req.addr, len_, req.data ) )

        s.left.resp.en  = Bits1(1)
        s.left.resp.msg = resp

    @s.update
    def up_memifc_rtl_fl_rdy():
      s.left.req.rdy = s.left.resp.rdy

class MemIfcFL2RTLAdapter( Component ):

  def construct( s, ReqType, RespType ):
    s.left  = MemMinionIfcFL ()
    s.right = MemMasterIfcRTL( ReqType, RespType )

    s.fl2cl       = MemIfcFL2CLAdapter( ReqType, RespType )
    s.req_cl2rtl  = RecvCL2SendRTL( ReqType )
    s.resp_rtl2cl = RecvRTL2SendCL( RespType)
    connect( s.left, s.fl2cl.left )
    connect_pairs(
      s.fl2cl.right.req, s.req_cl2rtl.recv,
      s.req_cl2rtl.send, s.right.req,
    )
    connect_pairs(
      s.fl2cl.right.resp, s.resp_rtl2cl.send,
      s.resp_rtl2cl.recv, s.right.resp,
    )



#=========================================================================
# TranslationImportPass.py
#=========================================================================
# Translate and import components having the `sverilog_translate_import`
# attribute.
#
# Author : Peitian Pan
# Date   : Aug 6, 2019

from pymtl3.passes.BasePass import BasePass

from .import_.ImportConfigs import ImportConfigs
from .import_.ImportPass import ImportPass
from .translation.TranslationPass import TranslationPass


class TranslationImportPass( BasePass ):

  def __call__( s, top ):
    s.top = top
    s.traverse_hierarchy( top )
    top.apply( s.get_translation_pass() )
    return s.get_import_pass()( top )

  def traverse_hierarchy( s, m ):
    if hasattr(m, s.get_flag_name()):
      if not hasattr(m, s.get_translation_flag_name()):
        setattr(m, s.get_translation_flag_name(), s.get_translation_configs())
      if not hasattr(m, s.get_import_flag_name()):
        setattr(m, s.get_import_flag_name(), s.get_import_configs())
    for child in m.get_child_components():
      s.traverse_hierarchy( child )

  def get_translation_pass( s ):
    return TranslationPass()

  def get_import_pass( s ):
    return ImportPass()

  def get_flag_name( s ):
    return "sverilog_translate_import"

  def get_translation_flag_name( s ):
    return "sverilog_translate"

  def get_import_flag_name( s ):
    return "config_sverilog_import"

  def get_translation_configs( s ):
    return True

  def get_import_configs( s ):
    return ImportConfigs(vl_Wno_list=['UNOPTFLAT', 'UNSIGNED'])



from .import_.ImportConfigs import ImportConfigs
from .import_.ImportPass import ImportPass
from .translation.TranslationPass import TranslationPass
from .TranslationImportPass import TranslationImportPass



#=========================================================================
# StructuralRTLIRGenL4Pass_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 19, 2019
"""Test the generation of level 1 structural RTLIR."""

from pymtl3.passes.rtlir.structural.StructuralRTLIRGenL4Pass import (
    StructuralRTLIRGenL4Pass,
)
from pymtl3.passes.rtlir.structural.StructuralRTLIRSignalExpr import *
from pymtl3.passes.testcases import (
    CaseBits32ArrayConnectSubCompAttrComp,
    CaseBits32ConnectSubCompAttrComp,
)

from .StructuralRTLIRGenL1Pass_test import gen_connections


def test_L4_subcomp_attr():
  a = CaseBits32ConnectSubCompAttrComp.DUT()
  a.elaborate()
  a.apply( StructuralRTLIRGenL4Pass( gen_connections( a ) ) )
  ns = a._pass_structural_rtlir_gen
  comp = CurComp(a, 's')
  # The first two signals are clk and reset
  assert ns.connections[2] == \
    (SubCompAttr(CurCompAttr(comp, 'b'), 'out'), CurCompAttr(comp, 'out'))

def test_L4_subcomp_index():
  a = CaseBits32ArrayConnectSubCompAttrComp.DUT()
  a.elaborate()
  a.apply( StructuralRTLIRGenL4Pass( gen_connections( a ) ) )
  ns = a._pass_structural_rtlir_gen
  comp = CurComp(a, 's')
  # The first ten signals are clks and resets
  assert ns.connections[10] == \
    (SubCompAttr(ComponentIndex(CurCompAttr(comp, 'b'), 1), 'out'),
      CurCompAttr(comp, 'out'))



"""
========================================================================
ComponentAPI_test.py
========================================================================

Author : Shunning Jiang
Date   : June 2, 2019
"""
import random

from pymtl3.datatypes import *
from pymtl3.dsl import Component, InPort, OutPort, Placeholder, Wire, connect
from pymtl3.dsl.errors import InvalidAPICallError

from .sim_utils import simple_sim_pass


def test_api_not_elaborated():

  class X( Component ):
    def construct( s, nbits=0 ):
      s.in_ = InPort ( mk_bits(nbits) )
      s.out = OutPort( mk_bits(nbits) )
      @s.update
      def up_x():
        s.out = s.in_ + 1

  class Y( Component ):
    def construct( s, nbits=0 ):
      s.in_ = InPort ( mk_bits(nbits) )
      s.out = OutPort( mk_bits(nbits) )
      s.x = X()( in_ = s.in_, out = s.out )

  a = Y()
  a.elaborate()
  try:
    a.x.get_all_update_blocks()
  except InvalidAPICallError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown InvalidAPICallError.")

# The following two tests cases test x.replace_component()

class Foo_shamt( Placeholder, Component ):
  def construct( s, shamt=1 ):
    s.in_ = InPort ( Bits32 )
    s.out = OutPort( Bits32 )

    # Nothing here

  def line_trace( s ):
    return "{}>{}".format( s.in_, s.out )

class Real_shamt( Component ):
  def construct( s, shamt=1 ):
    s.in_ = InPort ( Bits32 )
    s.out = OutPort( Bits32 )
    @s.update
    def up_real():
      s.out = s.in_ << shamt

  def line_trace( s ):
    return "{}>{}".format( s.in_, s.out )

class Real_shamt2( Component ):
  def construct( s, shamt=1 ):
    s.in_ = InPort ( Bits32 )
    s.out = OutPort( Bits32 )
    @s.update
    def up_real():
      s.out = s.in_ + shamt

  def line_trace( s ):
    return "{}>{}".format( s.in_, s.out )

class Foo_shamt_list_wrap( Component ):
  def construct( s, nbits=0 ):
    s.in_ = InPort ( mk_bits(nbits) )
    s.out = [ OutPort( mk_bits(nbits) ) for i in range(5) ]

    s.inner = [ Foo_shamt( i )( in_ = s.in_, out = s.out[i] ) for i in range(5) ]

  def line_trace( s ):
    return "|".join( [ x.line_trace() for x in s.inner ] )

def test_real_replaced_by_real2():

  class Real_wrap( Component ):
    def construct( s, nbits=0 ):
      s.in_ = InPort ( mk_bits(nbits) )
      s.out = OutPort( mk_bits(nbits) )
      s.w   = Wire( mk_bits(nbits) )
      connect( s.w, s.out )

      s.inner = Real_shamt( 5 )( in_ = s.in_, out = s.w )

    def line_trace( s ):
      return s.inner.line_trace()

  foo_wrap = Real_wrap( 32 )

  foo_wrap.elaborate()
  foo_wrap.replace_component( foo_wrap.inner, Real_shamt2, check=True )

  simple_sim_pass( foo_wrap )
  foo_wrap.sim_reset()

  foo_wrap.in_ = Bits32(100)
  foo_wrap.tick()
  print(foo_wrap.line_trace())
  assert foo_wrap.out == 105

  foo_wrap.in_ = Bits32(3)
  foo_wrap.tick()
  print(foo_wrap.line_trace())
  assert foo_wrap.out == 8

def test_replace_component_list_of_foo_by_real():

  foo_wrap = Foo_shamt_list_wrap( 32 )

  foo_wrap.elaborate()
  order = list(range(5))
  random.shuffle( order )
  for i in order:
    foo_wrap.replace_component( foo_wrap.inner[i], Real_shamt )

  simple_sim_pass( foo_wrap )

  print()
  foo_wrap.in_ = Bits32(16)
  foo_wrap.tick()
  print(foo_wrap.line_trace())

  foo_wrap.in_ = Bits32(4)
  foo_wrap.tick()
  print(foo_wrap.line_trace())

def test_replace_component_list_of_real_by_real2():

  foo_wrap = Foo_shamt_list_wrap( 32 )

  order = list(range(5))
  random.shuffle( order )

  foo_wrap.elaborate()

  for i in order:
    foo_wrap.replace_component( foo_wrap.inner[i], Real_shamt )

  random.shuffle( order )
  for i in order:
    foo_wrap.replace_component( foo_wrap.inner[i], Real_shamt2 )

  print(len(foo_wrap._dsl.connect_order))
  simple_sim_pass( foo_wrap )

  print()
  foo_wrap.in_ = Bits32(16)
  foo_wrap.tick()
  print(foo_wrap.line_trace())

  foo_wrap.in_ = Bits32(4)
  foo_wrap.tick()
  print(foo_wrap.line_trace())

def test_replace_deep_list_of_foo_by_real():

  class TopWrap( Component ):
    def construct( s ):
      s.in_    = InPort( Bits32 )
      s.foobar = Foo_shamt_list_wrap( 32 )( in_ = s.in_ )

  foo_wrap = TopWrap()

  foo_wrap.elaborate()
  order = list(range(5))
  random.shuffle( order )
  for i in order:
    foo_wrap.replace_component( foo_wrap.foobar.inner[i], Real_shamt )

  simple_sim_pass( foo_wrap )

  print()
  foo_wrap.in_ = Bits32(16)
  foo_wrap.tick()
  print(foo_wrap.foobar.line_trace())

  foo_wrap.in_ = Bits32(4)
  foo_wrap.tick()
  print(foo_wrap.foobar.line_trace())

# This test is to test if we save the fact that up_in writes the inport
# of the inner module and recover it when we put in the new component

def test_replace_component_upblk_rw_port():

  class Real_wrap( Component ):
    def construct( s, nbits=0 ):
      s.in_ = InPort ( mk_bits(nbits) )
      s.out = OutPort( mk_bits(nbits) )
      s.w   = Wire( mk_bits(nbits) )
      connect( s.w, s.out )

      @s.update
      def up_in():
        s.inner.in_ = s.in_

      @s.update
      def up_out():
        s.w = s.inner.out

      s.inner = Real_shamt( 5 )#( in_ = s.in_, out = s.w )

    def line_trace( s ):
      return s.inner.line_trace()

  foo_wrap = Real_wrap( 32 )

  foo_wrap.elaborate()
  foo_wrap.replace_component( foo_wrap.inner, Real_shamt2, check=True )

  simple_sim_pass( foo_wrap )
  foo_wrap.sim_reset()

  foo_wrap.in_ = Bits32(100)
  foo_wrap.tick()
  print(foo_wrap.line_trace())
  assert foo_wrap.out == 105

  foo_wrap.in_ = Bits32(3)
  foo_wrap.tick()
  print(foo_wrap.line_trace())
  assert foo_wrap.out == 8


# This test is the extended version of the previous one which tests if we
# save and recover functions in the parent correctly

def test_replace_component_func_rw_port():

  class Real_wrap( Component ):
    def construct( s, nbits=0 ):
      s.in_ = InPort ( mk_bits(nbits) )
      s.out = OutPort( mk_bits(nbits) )
      s.w   = Wire( mk_bits(nbits) )
      connect( s.w, s.out )

      @s.func
      def assign_in( x ):
        s.inner.in_ = x

      @s.update
      def up_in():
        assign_in( s.in_ )

      @s.func
      def read_out():
        s.w = s.inner.out

      @s.update
      def up_out():
        read_out()

      s.inner = Real_shamt( 5 )#( in_ = s.in_, out = s.w )

    def line_trace( s ):
      return s.inner.line_trace()

  foo_wrap = Real_wrap( 32 )

  foo_wrap.elaborate()
  foo_wrap.replace_component( foo_wrap.inner, Real_shamt2, check=True )

  simple_sim_pass( foo_wrap )
  foo_wrap.sim_reset()

  foo_wrap.in_ = Bits32(100)
  foo_wrap.tick()
  print(foo_wrap.line_trace())
  assert foo_wrap.out == 105

  foo_wrap.in_ = Bits32(3)
  foo_wrap.tick()
  print(foo_wrap.line_trace())
  assert foo_wrap.out == 8

def test_ctrl_dpath_connected_replaced_both():

  class Module1( Component ):
    def construct( s ):
      s.in_  = InPort( Bits32 )
      s.wire = Wire( Bits32 )
      s.out  = OutPort( Bits32 )

      connect( s.in_, s.wire )

      @s.update
      def out():
        s.out = s.wire + 10

  class Module2( Component ):
    def construct( s ):
      s.in_  = InPort( Bits32 )
      s.wire = Wire( Bits32 )
      s.out  = OutPort( Bits32 )

      connect( s.in_, s.wire )

      @s.update
      def out():
        s.out = s.wire + 444

  class Inner( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.m1  = Module1()
      s.m2  = Module1()
      s.out = OutPort( Bits32 )

      connect( s.in_, s.m1.in_ )
      connect( s.m1.out, s.m2.in_ )
      connect( s.m2.out, s.out )

  class Top( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.inner = Inner()
      s.out = OutPort( Bits32 )

      connect( s.in_, s.inner.in_)
      connect( s.inner.out, s.out )

    def line_trace( s ):
      return "{} > {}".format( s.in_, s.out )

  a = Top()

  a.elaborate()
  a.replace_component( a.inner.m1, Module2() )
  a.replace_component( a.inner.m2, Module2() )

  simple_sim_pass( a )

  a.in_ = Bits32(10)
  a.tick()
  assert a.out == 10 + 444 * 2

# Test orders

def test_connect_upblk_orders():

  class Inner( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      s.out2 = OutPort( Bits32 )
      @s.update
      def up_out():
        s.out = Bits32(0)

      s.wire = Wire(Bits32)
      @s.update_ff
      def up_ff():
        s.wire <<= s.wire + 1

      @s.update
      def up_out2():
        s.out2 = Bits32(0)

  class Top( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.inner = Inner()
      s.out = OutPort( Bits32 )
      connect( s.inner.out, s.out )
      connect( s.in_, s.inner.in_)

  a = Top()
  a.elaborate()
  assert a.get_connect_order() == [ (a.inner.clk, a.clk), (a.inner.reset, a.reset),
                                    (a.inner.out, a.out), (a.in_, a.inner.in_) ]

  assert not a.get_update_block_order()
  u = a.inner.get_update_block_order()
  assert u[0].__name__ == "up_out"
  assert u[1].__name__ == "up_ff"
  assert u[2].__name__ == "up_out2"

# def test_garbage_collection():

  # class X( Component ):
    # def construct( s, nbits=0 ):
      # s.leak = bytearray(1<<30)
      # s.in_ = InPort ( mk_bits(nbits) )
      # s.out = OutPort( mk_bits(nbits) )
      # @s.update
      # def up_x():
        # s.out = s.in_ + 1
  # for i in range(100):
    # x = X()
    # x = X()



"""
#=========================================================================
# LineTraceParamPass_test.py
#=========================================================================
# Test for multi-level line trace pass.
#
# Author : Yanghui Ou
#   Date : 29 May, 2019
"""
from pymtl3.dsl import *

from ..LineTraceParamPass import LineTraceParamPass


def test_multi_level_trace():

  class Top( Component ):
    def construct( s ):
      s.simple_trace = 'simple'
      s.verbose_trace = 'verbose'

    def line_trace( s, level='simple' ):
      if level == 'simple':
        return f"{s.simple_trace}"
      elif level == 'verbose':
        return f"{s.verbose_trace}"
      else:
        return "default"

  A = Top()
  A.set_param( 'top.line_trace', level='verbose' )
  A.elaborate()
  A.apply( LineTraceParamPass()      )
  A.lock_in_simulation()
  print( A.line_trace() )
  assert A.line_trace() == "verbose"



"""
========================================================================
queues.py
========================================================================
This file contains cycle-level queues.

Author : Shunning Jiang, Yanghui Ou
Date   : Mar 10, 2018
"""

from collections import deque

from pymtl3 import *

#-------------------------------------------------------------------------
# PipeQueueCL
#-------------------------------------------------------------------------

class PipeQueueCL( Component ):

  def construct( s, num_entries=1 ):
    s.queue = deque( maxlen=num_entries )

    s.add_constraints(
      M( s.peek   ) < M( s.enq  ),
      M( s.deq    ) < M( s.enq  )
    )

  @non_blocking( lambda s: len( s.queue ) < s.queue.maxlen )
  def enq( s, msg ):
    s.queue.appendleft( msg )

  @non_blocking( lambda s: len( s.queue ) > 0 )
  def deq( s ):
    return s.queue.pop()

  @non_blocking( lambda s: len( s.queue ) > 0 )
  def peek( s ):
    return s.queue[-1]

  def line_trace( s ):
    return "{}( ){}".format( s.enq, s.deq )

#-------------------------------------------------------------------------
# BypassQueueCL
#-------------------------------------------------------------------------

class BypassQueueCL( Component ):

  def construct( s, num_entries=1 ):
    s.queue = deque( maxlen=num_entries )

    s.add_constraints(
      M( s.enq    ) < M( s.peek    ),
      M( s.enq    ) < M( s.deq     ),
    )

  @non_blocking( lambda s: len( s.queue ) < s.queue.maxlen )
  def enq( s, msg ):
    s.queue.appendleft( msg )

  @non_blocking( lambda s: len( s.queue ) > 0 )
  def deq( s ):
    return s.queue.pop()

  @non_blocking( lambda s: len( s.queue ) > 0 )
  def peek( s ):
    return s.queue[-1]

  def line_trace( s ):
    return "{}( ){}".format( s.enq, s.deq )

#-------------------------------------------------------------------------
# NormalQueueCL
#-------------------------------------------------------------------------

class NormalQueueCL( Component ):

  def construct( s, num_entries=1 ):
    s.queue = deque( maxlen=num_entries )
    s.enq_rdy = False
    s.deq_rdy = False

    @s.update
    def up_pulse():
      s.enq_rdy    = len( s.queue ) < s.queue.maxlen
      s.deq_rdy    = len( s.queue ) > 0

    s.add_constraints(
      U( up_pulse ) < M( s.enq.rdy ),
      U( up_pulse ) < M( s.deq.rdy ),
      M( s.peek   ) < M( s.deq.rdy  ),
      M( s.peek   ) < M( s.enq.rdy  )
    )

  @non_blocking( lambda s: s.enq_rdy )
  def enq( s, msg ):
    s.queue.appendleft( msg )

  @non_blocking( lambda s: s.deq_rdy )
  def deq( s ):
    return s.queue.pop()

  @non_blocking( lambda s: len( s.queue ) > 0 )
  def peek( s ):
    return s.queue[-1]

  def line_trace( s ):
    return "{}( ){}".format( s.enq, s.deq )



#=========================================================================
# SVTranslator_L4_cases_test.py
#=========================================================================
"""Test the SystemVerilog translator."""

import pytest

from pymtl3.passes.backends.sverilog.util.test_utility import check_eq
from pymtl3.passes.rtlir.util.test_utility import get_parameter

from ..behavioral.test.SVBehavioralTranslatorL5_test import test_sverilog_behavioral_L5
from ..structural.test.SVStructuralTranslatorL4_test import test_sverilog_structural_L4
from ..SVTranslator import SVTranslator


def run_test( case, m ):
  m.elaborate()
  tr = SVTranslator( m )
  tr.translate( m )
  check_eq( tr.hierarchy.src, case.REF_SRC )

@pytest.mark.parametrize(
  'case', get_parameter('case', test_sverilog_behavioral_L5) + \
          get_parameter('case', test_sverilog_structural_L4)
)
def test_sverilog_L4( case ):
  run_test( case, case.DUT() )



#=========================================================================
# BehavioralRTLIRPass_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : Feb 2, 2019
"""Test the behavioral RTLIR passes.

The generation, type check, and visualization passes are invoked. The
results of generation pass are verifed against a reference AST.
"""

# from pymtl3.datatypes import *
from pymtl3.passes.rtlir.behavioral import (
    BehavioralRTLIRGenPass,
    BehavioralRTLIRTypeCheckPass,
    BehavioralRTLIRVisualizationPass,
)
from pymtl3.passes.rtlir.behavioral.BehavioralRTLIR import *
from pymtl3.passes.rtlir.behavioral.test.BehavioralRTLIRL1Pass_test import XFAIL_ON_PY3
from pymtl3.passes.rtlir.errors import PyMTLTypeError
from pymtl3.passes.rtlir.util.test_utility import do_test, expected_failure
from pymtl3.passes.testcases import (
    CaseBits8Bits16MismatchAssignComp,
    CaseBits16ConstAddComp,
    CaseBits16IndexBasicComp,
    CaseBits32Bits64SlicingBasicComp,
    CaseConstSizeSlicingComp,
    CaseFlipFlopAdder,
    CaseForBasicComp,
    CaseIfBasicComp,
    CaseReducesInx3OutComp,
    CaseSlicingOverIndexComp,
    CaseSubCompAddComp,
    CaseTwoUpblksSliceComp,
)


def local_do_test( m ):
  """Check if generated behavioral RTLIR is the same as reference."""

  ref = m._rtlir_test_ref
  m.elaborate()
  m.apply( BehavioralRTLIRGenPass() )
  m.apply( BehavioralRTLIRTypeCheckPass() )
  m.apply( BehavioralRTLIRVisualizationPass() )

  for blk in m.get_update_blocks():
    assert\
      m._pass_behavioral_rtlir_gen.rtlir_upblks[ blk ] == ref[ blk.__name__ ]

def test_reduce( do_test ):
  a = CaseReducesInx3OutComp.DUT()
  in_1 = Attribute( Base( a ), 'in_1' )
  in_2 = Attribute( Base( a ), 'in_2' )
  in_3 = Attribute( Base( a ), 'in_3' )
  out = Attribute( Base( a ), 'out' )

  a._rtlir_test_ref = { 'v_reduce' : CombUpblk( 'v_reduce', [
    Assign( out, BinOp(
      BinOp( Reduce( BitAnd(), in_1 ), BitAnd(), Reduce( BitOr(), in_2 ) ),
      BitOr(), Reduce( BitXor(), in_3 ),
    ), True )
  ] ) }

  # a._test_vector = [
  #         ' in_1        in_2     in_3     *out',
  #   [     Bits32,    Bits32,   Bits32,   Bits1    ],

  #   [          0,         1,        2,        1   ],
  #   [    b32(-1),         1,  b32(-1),        1   ],
  #   [          9,         8,        7,        1   ],
  # ]

  do_test( a )

def test_index_basic( do_test ):
  a = CaseBits16IndexBasicComp.DUT()

  a._rtlir_test_ref = { 'index_basic' : CombUpblk( 'index_basic', [
    Assign( Index( Attribute( Base( a ), 'out' ), Number( 0 ) ),
      BinOp( Index( Attribute( Base( a ), 'in_' ), Number( 0 ) ), Add(),
             Index( Attribute( Base( a ), 'in_' ), Number( 1 ) ) ), True ),
    Assign( Index( Attribute( Base( a ), 'out' ), Number( 1 ) ),
      BinOp( Index( Attribute( Base( a ), 'in_' ), Number( 2 ) ), Add(),
             Index( Attribute( Base( a ), 'in_' ), Number( 3 ) ) ), True )
  ] ) }

  # a._test_vector = [
  #         'in_[0]     in_[1]    in_[2]    in_[3]    *out[0]     *out[1]',
  #   [     Bits16,    Bits16,   Bits16,   Bits16,    Bits16,     Bits16 ],

  #   [          0,         1,        2,        3,         1,          5 ],
  #   [ Bits16(-1),         1, Bits16(-1),      1,         0,          0 ],
  #   [          9,         8,        7,        6,        17,         13 ],
  # ]

  do_test( a )

def test_mismatch_width_assign( do_test ):
  a = CaseBits8Bits16MismatchAssignComp.DUT()

  a._rtlir_test_ref = { 'mismatch_width_assign' : CombUpblk(
    'mismatch_width_assign', [ Assign(
      Attribute( Base( a ), 'out' ), Attribute( Base( a ), 'in_' ), True
    )
  ] ) }

  # a._test_vector = [
  #               'in_             *out',
  #   [        Bits16,           Bits8 ],

  #   [             0,               0 ],
  #   [             2,               2 ],
  #   [    Bits16(-1),       Bits8(-1) ],
  #   [    Bits16(-2),       Bits8(-2) ],
  #   [ Bits16(32767),      Bits8(255) ],
  # ]

  with expected_failure( PyMTLTypeError, "LHS and RHS of assignment" ):
    do_test( a )

def test_slicing_basic( do_test ):
  a = CaseBits32Bits64SlicingBasicComp.DUT()

  a._rtlir_test_ref = { 'slicing_basic' : CombUpblk( 'slicing_basic', [
    Assign( Slice( Attribute( Base( a ), 'out' ), Number( 0 ), Number( 16 ) ),
      Slice( Attribute( Base( a ), 'in_' ), Number( 16 ), Number( 32 ) ), True ),
    Assign( Slice( Attribute( Base( a ), 'out' ), Number( 16 ), Number( 32 ) ),
      Slice( Attribute( Base( a ), 'in_' ), Number( 0 ), Number( 16 ) ), True )
  ] ) }

  # a._test_vector = [
  #               'in_                        *out',
  #   [        Bits32,                     Bits64 ],

  #   [             0,                          0 ],
  #   [             2,            Bits64(0x20000) ],
  #   [    Bits32(-1),         Bits64(0xffffffff) ],
  #   [    Bits32(-2),         Bits64(0xfffeffff) ],
  #   [ Bits32(32767),         Bits64(0x7fff0000) ],
  # ]

  do_test( a )

def test_bits_basic( do_test ):
  a = CaseBits16ConstAddComp.DUT()

  a._rtlir_test_ref = { 'bits_basic' : CombUpblk( 'bits_basic', [
    Assign( Attribute( Base( a ), 'out' ),
      BinOp( Attribute( Base( a ), 'in_' ), Add(), SizeCast( 16, Number( 10 ) ) ), True )
  ] ) }

  # a._test_vector = [
  #               'in_              *out',
  #   [        Bits16,           Bits16 ],
  #   [             0,               10 ],
  #   [             2,               12 ],
  #   [    Bits16(-1),        Bits16(9) ],
  #   [    Bits16(-2),        Bits16(8) ],
  #   [Bits16(0x7FFF),   Bits16(0x8009) ],
  # ]

  do_test( a )

def test_index_bits_slicing( do_test ):
  a = CaseSlicingOverIndexComp.DUT()

  a._rtlir_test_ref = { 'index_bits_slicing' : CombUpblk( 'index_bits_slicing', [
    Assign( Slice(
      Index( Attribute( Base( a ), 'out' ), Number( 0 ) ),
      Number( 0 ), Number( 8 )
      ),
      BinOp(
        BinOp(
          Slice( Index( Attribute( Base( a ), 'in_' ), Number( 1 ) ), Number( 8 ), Number( 16 ) ),
          Add(),
          Slice( Index( Attribute( Base( a ), 'in_' ), Number( 2 ) ), Number( 0 ), Number( 8 ) ),
        ),
        Add(),
        SizeCast( 8, Number( 10 ) )
      ),
      True
    ),
    Assign(
      Index( Attribute( Base( a ), 'out' ), Number( 1 ) ),
      BinOp(
        BinOp(
          Slice( Index( Attribute( Base( a ), 'in_' ), Number( 3 ) ), Number( 0 ), Number( 16 ) ),
          Add(),
          Index( Attribute( Base( a ), 'in_' ), Number( 4 ) )
        ),
        Add(),
        SizeCast( 16, Number( 1 ) )
      ),
      True
    ),
  ] ) }

  # a._test_vector = [
  #     'in_[0] in_[1] in_[2] in_[3] in_[4] in_[5] in_[6] in_[7] in_[8] in_[9]\
  #         *out[0] *out[1] *out[2] *out[3] *out[4]',
  #   [ Bits16 ] * 15,

  #   # 8-bit truncation!
  #   [ Bits16(0xff) ] * 10 + [ Bits16(0x09), Bits16(0x01ff), 0, 0, 0 ],
  #   [ Bits16(0x00) ] * 10 + [ Bits16(0x0a), Bits16(0x0001), 0, 0, 0 ],
  # ]

  do_test( a )

def test_multi_components( do_test ):
  a = CaseSubCompAddComp.DUT()

  a._rtlir_test_ref = { 'multi_components_A' : CombUpblk( 'multi_components_A', [
    Assign( Attribute( Base( a ), 'out' ),
      BinOp(
        Attribute( Base( a ), 'in_' ),
        Add(),
        Attribute( Attribute( Base( a ), 'b' ), 'out' )
      ),
      True
    )
  ] ) }

  # a._test_vector = [
  #               'in_              *out',
  #   [ Bits16 ] *2,

  #   [             0,               0 ],
  #   [             2,               4 ],
  #   [    Bits16(-1),      Bits16(-2) ],
  #   [    Bits16(-2),      Bits16(-4) ],
  # ]

  do_test( a )

def test_if_basic( do_test ):
  a = CaseIfBasicComp.DUT()

  a._rtlir_test_ref = {
    'if_basic' : CombUpblk( 'if_basic', [ If(
      Compare( Slice( Attribute( Base( a ), 'in_' ), Number( 0 ), Number( 8 ) ), Eq(), SizeCast( 8, Number( 255 ) ) ),
      [ Assign( Attribute( Base( a ), 'out' ), Slice( Attribute( Base( a ), 'in_' ), Number( 8 ), Number( 16 ) ), True ) ],
      [ Assign( Attribute( Base( a ), 'out' ), SizeCast( 8, Number( 0 ) ), True ) ]
    )
  ] ) }

  # a._test_vector = [
  #               'in_              *out',
  #   [         Bits16,           Bits8 ],
  #   [           255,                0 ],
  #   [           511,                1 ],
  #   [           256,                0 ],
  # ]

  do_test( a )

@XFAIL_ON_PY3
def test_for_basic( do_test ):
  a = CaseForBasicComp.DUT()

  twice_i = BinOp( Number( 2 ), Mult(), LoopVar( 'i' ) )

  a._rtlir_test_ref = {
    'for_basic' : CombUpblk( 'for_basic', [ For(
      LoopVarDecl( 'i' ), Number( 0 ), Number( 8 ), Number( 1 ),
      [ Assign(
          Slice( Attribute( Base( a ), 'out' ), twice_i, BinOp( twice_i, Add(), Number( 1 ) ) ),
          BinOp(
            Slice( Attribute( Base( a ), 'in_' ), twice_i, BinOp( twice_i, Add(), Number( 1 ) ) ),
            Add(),
            Slice( Attribute( Base( a ), 'in_' ),
              BinOp( twice_i, Add(), Number( 1 ) ),
              BinOp( BinOp( twice_i, Add(), Number( 1 ) ), Add(), Number( 1 ) )
            )
          ),
          True
        )
      ]
    ) ] )
  }

  # a._test_vector = []

  do_test( a )

def test_multi_upblks( do_test ):
  a = CaseTwoUpblksSliceComp.DUT()

  a._rtlir_test_ref = { 'multi_upblks_1' : CombUpblk( 'multi_upblks_1', [
      Assign( Slice( Attribute( Base( a ), 'out' ), Number(0), Number(4) ), Attribute( Base( a ), 'in_' ), True ),
    ] ),
    'multi_upblks_2' : CombUpblk( 'multi_upblks_2', [
      Assign( Slice( Attribute( Base( a ), 'out' ), Number(4), Number(8) ), Attribute( Base( a ), 'in_' ), True ),
    ] )
  }

  # a._test_vector = [
  #               'in_              *out',
  #   [         Bits4,            Bits8 ],

  #   [     Bits4(-1),      Bits8(0xff) ],
  #   [      Bits4(1),      Bits8(0x11) ],
  #   [      Bits4(7),      Bits8(0x77) ],
  # ]

  do_test( a )

def test_ff_upblk( do_test ):
  a = CaseFlipFlopAdder.DUT()
  a._rtlir_test_ref = {
      'update_ff_upblk' : SeqUpblk( 'update_ff_upblk', [
        Assign( Attribute( Base( a ), 'out0' ), BinOp( Attribute( Base( a ), 'in0' ), Add(), Attribute( Base( a ), 'in1' ) ), False ),
        ] )
  }
  # a._test_vector = [
  #               'in0                in1              *out',
  #   [         Bits32,            Bits32,            Bits32 ],

  #   [     Bits32(-1),       Bits32(0xff),     Bits32(0xfe) ],
  #   [     Bits32(1),        Bits32(0x11),     Bits32(0x12) ],
  #   [     Bits32(7),        Bits32(0x77),     Bits32(0x7d) ],
  # ]

  do_test( a )

def test_fixed_size_slice( do_test ):
  a = CaseConstSizeSlicingComp.DUT()
  a._rtlir_test_ref = {
      'upblk' : CombUpblk( 'upblk', [
        For( LoopVarDecl('i'), Number(0), Number(2), Number(1), [
          Assign(
            Index(Attribute(Base(a), 'out'), LoopVar('i')),
            Slice(Attribute(Base(a), 'in_'),
                  BinOp(LoopVar('i'), Mult(), Number(8)),
                  BinOp(BinOp(LoopVar('i'), Mult(), Number(8)), Add(), Number(8)),
                  BinOp(LoopVar('i'), Mult(), Number(8)),
                  8,
            ), True)]
        )
      ])
  }
  # a._test_vector = [
  #               'in_              *out[0]          *out[1]',
  #   [         Bits16,               Bits8,            Bits8 ],

  #   [     Bits16(-1),         Bits8(0xff),      Bits8(0xff) ],
  #   [      Bits16(1),         Bits8(0x01),      Bits8(0x00) ],
  #   [      Bits16(7),         Bits8(0x07),      Bits8(0x00) ],
  #   [ Bits16(0xff00),         Bits8(0x00),      Bits8(0xff) ],
  #   [ Bits16(0x3412),         Bits8(0x12),      Bits8(0x34) ],
  #   [ Bits16(0x9876),         Bits8(0x76),      Bits8(0x98) ],
  # ]

  do_test( a )



#=========================================================================
# YosysBehavioralTranslatorL3.py
#=========================================================================
# Author : Peitian Pan
# Date   : June 9, 2019
"""Provide the yosys-compatible SystemVerilog L3 behavioral translator."""

from pymtl3.datatypes import Bits, is_bitstruct_inst
from pymtl3.passes.backends.sverilog.errors import SVerilogTranslationError
from pymtl3.passes.backends.sverilog.translation.behavioral.SVBehavioralTranslatorL3 import (
    BehavioralRTLIRToSVVisitorL3,
    SVBehavioralTranslatorL3,
)
from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir import RTLIRType as rt

from .YosysBehavioralTranslatorL2 import (
    YosysBehavioralRTLIRToSVVisitorL2,
    YosysBehavioralTranslatorL2,
)


class YosysBehavioralTranslatorL3(
    YosysBehavioralTranslatorL2, SVBehavioralTranslatorL3 ):

  def _get_rtlir2sv_visitor( s ):
    return YosysBehavioralRTLIRToSVVisitorL3

class YosysBehavioralRTLIRToSVVisitorL3(
    YosysBehavioralRTLIRToSVVisitorL2, BehavioralRTLIRToSVVisitorL3 ):
  """IR visitor that generates yosys-compatible SystemVerilog code.

  This visitor differs from the canonical SystemVerilog visitor in that
  it name-mangles each struct signal into multiple signals for all fields
  in the struct. We don't use SystemVerilog struct here because yosys
  does not support that yet.
  """

  def _literal_number( s, nbits, value ):
    value = int( value )
    return f"{nbits}'d{value}"

  def _struct_instance( s, dtype, struct ):
    def _gen_packed_array( dtype, n_dim, array ):
      if not n_dim:
        if isinstance( dtype, rdt.Vector ):
          return s._literal_number( dtype.nbits, array )
        elif isinstance( dtype, rdt.Struct ):
          return s._struct_instance( dtype, array )
        else:
          assert False, f"unrecognized data type {Type}!"
      else:
        ret = []
        for i in reversed( range( n_dim[0]) ):
          ret.append( _gen_packed_array( dtype, n_dim[1:], array[i] ) )
        if n_dim[0] > 1:
          cat_str = "{" + ", ".join( ret ) + "}"
        else:
          cat_str = ", ".join( ret )
        return cat_str
    fields = []
    for name, Type in dtype.get_all_properties().items():
      field = getattr( struct, name )
      if isinstance( Type, rdt.Vector ):
        _field = s._literal_number( Type.nbits, field )
      elif isinstance( Type, rdt.Struct ):
        _field = s._struct_instance( Type, field )
      elif isinstance( Type, rdt.PackedArray ):
        n_dim = Type.get_dim_sizes()
        sub_dtype = Type.get_sub_dtype()
        _field = _gen_packed_array( sub_dtype, n_dim, field )
      else:
        assert False, f"unrecognized data type {Type}!"
      fields.append( _field )
    if len( fields ) == 1:
      struct_str = fields[0]
    else:
      struct_str = "{" + ", ".join( fields ) + "}"
    return struct_str

  #-----------------------------------------------------------------------
  # visit_Attribute
  #-----------------------------------------------------------------------

  def visit_Attribute( s, node ):
    """Return the SystemVerilog representation of an attribute.

    Add support for accessing struct attribute in L3.
    """
    if isinstance( node.value.Type, rt.Signal ):
      if isinstance( node.value.Type, rt.Const ):
        try:
          obj = node.Type.get_object()
        except AttributeError:
          obj = None
        if obj is None:
          raise SVerilogTranslationError( s.blk, node,
            f"attribute ({node.attr}) of constant struct instance ({node.value}) is not supported!" )
        else:
          if isinstance( obj, Bits ):
            s.signal_expr_prologue( node )
            node.sexpr['s_attr'] = \
                s._literal_number(obj.nbits, int(obj))
            node.sexpr["s_index"] = ""
            attr = node.attr
            return s.signal_expr_epilogue(node, f"{attr}")
          elif is_bitstruct_inst( obj ):
            s.signal_expr_prologue( node )
            node.sexpr['s_attr'] = \
                s._struct_instance(node.Type.get_dtype(), obj)
            node.sexpr["s_index"] = ""
            attr = node.attr
            return s.signal_expr_epilogue(node, f"{attr}")
          else:
            raise SVerilogTranslationError( s.blk, node,
              f"attribute ({node.attr}) of constant struct instance ({node.value}) is not supported!" )

      elif isinstance( node.value.Type.get_dtype(), rdt.Struct ):
        value = s.visit( node.value )
        s.signal_expr_prologue( node )
        attr = node.attr
        s.check_res( node, attr )
        node.sexpr['s_attr'] += "__{}"
        node.sexpr['attr'].append( attr )
        return s.signal_expr_epilogue(node, f"{value}.{attr}")

    return super().visit_Attribute( node )

  #-----------------------------------------------------------------------
  # visit_StructInst
  #-----------------------------------------------------------------------

  def visit_StructInst( s, node ):
    for value in node.values:
      value._top_expr = 1
    return super().visit_StructInst( node )



"""
========================================================================
errors.py
========================================================================

Author : Shunning Jiang
Date   : Jul 4, 2017
"""
class PassOrderError( Exception ):
  """ Raise when applying a pass to a component and some required variable
      generated by other passes is missing """
  def __init__( self, var ):
    return super().__init__( f"Please first apply other passes to generate model.{var}" )

class ModelTypeError( Exception ):
  """ Raise when a pass cannot be applied to some component type """
  def __init__( self, typename ):
    return super().__init__( f"This pass can only be applied to {typename}" )

class TranslationError( Exception ):
  """ Raise when translation goes wrong """
  def __init__( self, blk, x ):
    return super().__init__( f"{blk.__name__} {x}" )

class InvalidPassOption( Exception ):
  """ Raised when the given pass option is not valid. """
  def __init__( self, opt, pas ):
    return super().__init__(f"{opt} is not a valid option of pass {pas}!")

class InvalidPassOptionValue( Exception ):
  """ Raised when the given pass option value is not valid. """
  def __init__( self, opt, val, pas, msg ):
    return super().__init__(f"{val} is not a valid value for option {opt}"
                            f" of pass {pas} because {msg}.")



from ..BasePass import BasePass
from ..sim.GenDAGPass import GenDAGPass
from ..sim.SimpleSchedulePass import SimpleSchedulePass
from ..sim.WrapGreenletPass import WrapGreenletPass
from .HeuristicTopoPass import HeuristicTopoPass
from .TraceBreakingSchedTickPass import TraceBreakingSchedTickPass
from .UnrollTickPass import UnrollTickPass


class UnrollSim( BasePass ):
  def __call__( s, top ):
    top.elaborate()
    GenDAGPass()( top )
    WrapGreenletPass()( top )
    SimpleSchedulePass()( top )
    UnrollTickPass()( top )
    top.lock_in_simulation()

class HeuTopoUnrollSim( BasePass ):
  def __call__( s, top ):
    top.elaborate()
    GenDAGPass()( top )
    WrapGreenletPass()( top )
    HeuristicTopoPass()( top )
    UnrollTickPass()( top )
    top.lock_in_simulation()

class TraceBreakingSim( BasePass ):
  def __call__( s, top ):
    top.elaborate()
    GenDAGPass()( top )
    WrapGreenletPass()( top )
    TraceBreakingSchedTickPass()( top )
    top.lock_in_simulation()



"""
========================================================================
set_param_test.py
========================================================================

Author : Yanhui Ou, Cheng Tan
Date   : May 28, 2019
"""
from pymtl3.datatypes.bits_import import Bits4, Bits8
from pymtl3.dsl.Component import Component
from pymtl3.dsl.ComponentLevel3 import connect
from pymtl3.dsl.Connectable import InPort, OutPort
from pymtl3.dsl.NamedObject import NamedObject

from .sim_utils import simple_sim_pass


class Animal( NamedObject ):
  def construct( s, lunch="dirt", dinner="dirt" ):
    s.lunch  = lunch
    s.dinner = dinner

class AnimalHouse( NamedObject ):
  def construct( s, AnimalTypes=[] ):
    s.animals = [ A() for A in AnimalTypes ]

class HoneyBadger( Animal ):
  pass

class Dromaius( Animal ):
  pass

class Panda( Animal ):
  pass

class Tiger( Animal ):
  def construct( s, MeatTypes=[], lunch="dirt", dinner="dirt" ):
    s.meats  = [ Meat() for Meat in MeatTypes ]
    s.lunch  = lunch
    s.dinner = dinner

class BabyTiger( Tiger ):
  pass

class Aquarium( AnimalHouse ):
  pass

class TigerTerrace( AnimalHouse ):
  pass

class PandaHouse( AnimalHouse ):
  pass

class Zoo( NamedObject ):
  def construct( s, AnimalHouseTypes=[] ):
    s.houses = [ H() for H in AnimalHouseTypes ]

def test_simple():
  A = Dromaius()
  A.set_param( "top.construct", lunch="grass" )
  A.elaborate()

  print( A.lunch )
  assert A.lunch  == "grass"
  assert A.dinner == "dirt"

def test_set_param_overwrite():
  Z = Zoo()
  Z.set_param( "top.construct", AnimalHouseTypes=[ PandaHouse, TigerTerrace, AnimalHouse, Aquarium ] )
  Z.set_param( "top.houses[0].construct", AnimalTypes=[ Panda, Panda, Panda   ] )
  Z.set_param( "top.houses[1].construct", AnimalTypes=[ Tiger, BabyTiger      ] )
  Z.set_param( "top.houses[2].construct", AnimalTypes=[ Dromaius, HoneyBadger ] )
  Z.set_param( "top.houses[1].animals[0].construct", MeatTypes=[ Dromaius, HoneyBadger ] )
  Z.set_param( "top.houses[0].animals*.construct", lunch="bamboo", dinner="bamboo" )
  Z.set_param( "top.houses[1].animals[0].meats*.construct", lunch="bamboo" )
  Z.set_param( "top.houses[1].animals[0].meats[1].construct", lunch="grass" )
  # The following set_param test tricky overwrite behavior.
  Z.set_param( "top.houses[2].animals[0].construct", lunch="bugs", dinner="bugs" )
  Z.set_param( "top.houses[2].animals*.construct", lunch="grass" )
  Z.set_param( "top.houses[2].animals[0].construct", lunch="bugs" )
  Z.elaborate()

  assert isinstance( Z.houses[0], PandaHouse )
  assert isinstance( Z.houses[1], TigerTerrace )
  assert isinstance( Z.houses[2], AnimalHouse )
  assert isinstance( Z.houses[3], Aquarium )
  for i in range( 3 ):
    assert isinstance( Z.houses[0].animals[i], Panda )
    assert Z.houses[0].animals[i].lunch  == "bamboo"
    assert Z.houses[0].animals[i].dinner == "bamboo"

  assert isinstance( Z.houses[1].animals[0], Tiger )
  assert isinstance( Z.houses[1].animals[1], BabyTiger )
  assert Z.houses[1].animals[0].meats[0].lunch == "bamboo"
  assert Z.houses[1].animals[0].meats[1].lunch == "grass"

  assert isinstance( Z.houses[2].animals[0], Dromaius )
  assert isinstance( Z.houses[2].animals[1], HoneyBadger )
  # FIXME
  assert Z.houses[2].animals[0].lunch  == "bugs"
  assert Z.houses[2].animals[0].dinner == "bugs"
  assert Z.houses[2].animals[1].lunch  == "grass"
  assert Z.houses[2].animals[1].dinner == "dirt"

def test_multi_regex():
  Z = Zoo()
  Z.set_param( "top.construct", AnimalHouseTypes=[ PandaHouse, TigerTerrace, AnimalHouse, Aquarium ] )
  Z.set_param( "top.houses[0].construct", AnimalTypes=[ Panda, Panda, Panda   ] )
  Z.set_param( "top.houses[1].construct", AnimalTypes=[ Tiger, BabyTiger      ] )
  Z.set_param( "top.houses[2].construct", AnimalTypes=[ Dromaius, HoneyBadger ] )
  Z.set_param( "top.houses[0].animals[0].construct", lunch="potato", dinner="bamboo" )
  Z.set_param( "top.houses*.animals*.construct", lunch="onion" )
  Z.set_param( "top.houses*.animals[0].construct", lunch="potato" )
  Z.elaborate()

  assert Z.houses[0].animals[0].lunch  == "onion"
  assert Z.houses[0].animals[0].dinner == "bamboo"

def test_set_param_hierarchical():
  Z = Zoo()

  Z.set_param( "top.construct", AnimalHouseTypes=[ Aquarium, TigerTerrace ] )
  Z.set_param( "top.houses[1].construct", AnimalTypes=[ Tiger, Panda ] )
  Z.set_param( "top.houses*.animals*.construct", lunch="grass" )
  Z.set_param( "top.houses*.animals[1].construct", lunch="meat" )

  #TODO: Whether we should check throw an error in following scenario
  Z.set_param( "top.houses*.xxx*.construct", lunch="grass" )

  Z.elaborate()

  assert isinstance(Z.houses[1], TigerTerrace)
  assert Z.houses[1].animals[0].lunch == "grass"
  assert Z.houses[1].animals[1].lunch == "meat"

def test_component():
  class Incr( Component ):
    def construct( s, DataType=Bits4(), incr_value=1 ):
      s.in_ = InPort ( DataType )
      s.out = OutPort( DataType )

      s.incr_value = DataType( incr_value )

      @s.update
      def up_incr():
        s.out = s.in_ + s.incr_value

  class IncrArray( Component ):
    def construct( s, num_incrs=1, DataType=Bits4() ):
      s.in_ = [ InPort ( DataType ) for _ in range( num_incrs ) ]
      s.out = [ OutPort( DataType ) for _ in range( num_incrs ) ]

      s.incrs = [ Incr( DataType=DataType ) for _ in range( num_incrs ) ]

      for i in range( num_incrs ):
        connect( s.in_[i], s.incrs[i].in_ )
        connect( s.out[i], s.incrs[i].out )

  A = Incr()
  A.set_param( "top.construct", DataType=Bits8, incr_value=2 )
  A.elaborate()
  A.apply( simple_sim_pass )
  A.in_ = 1
  A.tick()
  print( "A.out==", A.out )
  assert A.out == 3

  B = IncrArray()
  B.set_param( "top.construct", DataType=Bits8, num_incrs=3 )
  B.set_param( "top.incrs*.construct", incr_value=3 )
  B.set_param( "top.incrs[2].construct", incr_value=5 )
  B.elaborate()
  B.apply( simple_sim_pass )
  B.in_[0] = 1
  B.in_[1] = 2
  B.in_[2] = 3
  B.tick()
  assert B.out[0] == 4
  assert B.out[1] == 5
  assert B.out[2] == 8



"""
-------------------------------------------------------------------------
TraceBreakingSchedTickPas.py
-------------------------------------------------------------------------
Generate the schedule and tick with trace breaking + heuristic toposort.

Author : Shunning Jiang
Date   : Dec 26, 2018
"""
import py

from pymtl3.dsl import *

from ..BasePass import BasePass, PassMetadata
from ..errors import PassOrderError
from ..sim.SimpleSchedulePass import SimpleSchedulePass, check_schedule
from .HeuristicTopoPass import CountBranches

# FIXME also apply branchiness to all update_ff blocks

class TraceBreakingSchedTickPass( BasePass ):
  def __call__( self, top ):
    if not hasattr( top._dag, "all_constraints" ):
      raise PassOrderError( "all_constraints" )

    top._sched = PassMetadata()

    self.meta_schedule( top )
    # Reuse simple's ff and flip schedule
    simple = SimpleSchedulePass()
    simple.schedule_ff( top )
    simple.schedule_posedge_flip( top )

    self.trace_breaking_tick( top )

  def meta_schedule( self, top ):

    # Construct the intra-cycle graph based on normal update blocks

    V   = top._dag.final_upblks - top.get_all_update_ff()
    E   = set()
    Es  = { v: [] for v in V }
    InD = { v: 0  for v in V }

    for (u, v) in top._dag.all_constraints: # u -> v
      if u in V and v in V:
        InD[v] += 1
        Es[u].append( v )
        E.add( (u, v) )

    # Extract branchiness

    # Initialize all generated net block to 0 branchiness
    branchiness = { x: 0 for x in top._dag.genblks }

    visitor = CountBranches()
    for blk in top.get_all_update_blocks():
      hostobj = top.get_update_block_host_component( blk )
      branchiness[ blk ] = visitor.enter( hostobj.get_update_block_info( blk )[-1] )

    # Shunning: now we make the scheduling aware of meta blocks
    # Basically we enhance the topological sort to choose
    # branchy/unbranchy block based on the progress of the current meta
    # block. If there is no branch at all along the meta block, I let it
    # grow as long as possible. If we have to append one branchy update
    # block, we then append a couple of branchy blocks till the
    # branchiness bound is reached, after which we break the trace.
    #
    # TODO now I'm using binary-search. Ideally a double-ended
    # priority queue or a balanced binary search tree (AVL/Treap/RB) can
    # reduce it to O(nlogn). I didn't find any efficient python impl.

    def insert_sortedlist( arr, priority, item ):
      left, right = 0, len(arr)
      while left < right-1:
        mid = (left + right) >> 1
        if priority <= arr[mid][0]:
          right = mid
        else:
          left = mid
      arr.insert( right, (priority, item) )

    Q = []
    for v in V:
      if not InD[v]:
        br = branchiness[ v ]
        insert_sortedlist( Q, br, v )

    # Branchiness factor is the bound of branchiness in a meta block.
    branchiness_factor = 8

    # Block factor is the bound of the number of branchy blocks in a
    # meta block.
    branchy_block_factor = 4

    schedule = []

    metas = []
    current_meta = []
    current_branchiness = current_blk_count = 0

    while Q:
      # If currently there is no branchiness, append less branchy block
      if current_branchiness == 0:
        (br, u) = Q.pop(0)

        # Update the current
        current_blk_count += 1
        current_branchiness += br
        current_meta.append( u )

        if current_branchiness >= branchiness_factor:
          metas.append( current_meta )
          current_branchiness = current_blk_count = 0
          current_meta = []

      # We already append a branchy block
      else:
        # Find the most branchy block
        (br, u) = Q.pop()

        # If no branchy block available, directly start a new metablock

        if not br:
          metas.append( current_meta )
          current_branchiness = current_blk_count = 0
          current_meta = [ u ]

        # Limit the number of branchiness and number of branchy blocks

        elif current_branchiness + br <= branchiness_factor:
          current_meta.append( u )
          current_branchiness += br
          current_blk_count += 1

          if current_blk_count >= branchy_block_factor:
            metas.append( current_meta )
            current_branchiness = current_blk_count = 0
            current_meta = []

        else:
          current_meta.append( u )
          metas.append( current_meta )
          current_branchiness = current_blk_count = 0
          current_meta = []

      schedule.append( u )
      for v in Es[u]:
        InD[v] -= 1
        if not InD[v]:
          insert_sortedlist( Q, branchiness[ v ], v )

    # Append the last meta block
    if current_meta:
      metas.append( current_meta )

    print("num_metablks:", len(metas))

    for meta in metas:
      print("---------------")
      for blk in meta:
        print(" - {}: {}".format( blk.__name__, branchiness[ blk ] ))

    top._sched.meta_schedule = metas
    self.branchiness = branchiness

    check_schedule( top, schedule, V, E, InD )

  def trace_breaking_tick( self, top ):

    metas = [ top._sched.schedule_ff + top._sched.schedule_posedge_flip ] + top._sched.meta_schedule

    # We will use pypyjit.dont_trace_here to disable tracing across
    # intermediate update blocks.
    gen_tick_src =  "try:\n"
    gen_tick_src += "  from pypyjit import dont_trace_here\n"
    gen_tick_src += "except ImportError:\n"
    gen_tick_src += "  pass\n"

    # The "comment" that will be used for update calls.
    schedule_names = {}

    for i in range( len(metas) ):
      meta = metas[i]
      for j in range( len(meta) ):
        blk = meta[j]
        schedule_names[ (i, j) ] = "[br: {}] {}" \
          .format( self.branchiness[ blk ] if blk in self.branchiness else "???",
                   blk.__name__ )

        # Copy the scheduled functions to update_blkX__Y
        gen_tick_src += "update_blk{0}__{1} = metas[{0}][{1}];".format( i, j )

    for i in range( len(metas) ):
      meta_blk = metas[i]

      gen_tick_src += "\n\ndef meta_blk{}():\n  ".format(i)

      gen_tick_src += "\n  ".join( [ "update_blk{}__{}() # {}" \
                                    .format( i, j, schedule_names[(i, j)] )
                                    for j in range( len(meta_blk) )] )

      if i < len(metas)-1:
        gen_tick_src += "\ntry:\n"
        gen_tick_src += "  dont_trace_here(0, False, meta_blk{}.__code__)\n".format( i )
        gen_tick_src += "except NameError:\n"
        gen_tick_src += "  pass\n"

    gen_tick_src += "\ndef tick_top():\n  "
    gen_tick_src += "; ".join( [ "meta_blk{}()".format(i) for i in range(len(metas)) ] )

    local = locals()
    exec(py.code.Source( gen_tick_src ).compile(), local )

    #  print gen_tick_src
    top.tick = local["tick_top"]



"""
=========================================================================
TestCase.py
=========================================================================
Implement the base TestCase class.

Author : Peitian Pan
Date   : Dec 12, 2019
"""

from textwrap import dedent


class AliasOf:
  def __init__( s, alias_name = 'A' ):
    s.alias_name = alias_name

  def __get__( s, instance, owner ):
    return getattr( owner, s.alias_name )

  def __set__( s, instance, value ):
    # Overwriting an existing alias is not supported yet
    raise NotImplementedError

def set_attributes( _cls, *args ):
  assert len( args ) % 2 == 0
  # Correctly "copy" a Python class
  # https://stackoverflow.com/questions/9541025/how-to-copy-a-python-class
  cls = type(_cls.__name__, (_cls,), dict(_cls.__dict__))
  for attr, obj in zip( args[::2], args[1::2] ):
    setattr( cls, attr, dedent(obj) if isinstance(obj, str) else obj )
  return cls



"""
========================================================================
ComponentLevel6_test.py
========================================================================
Tests for component level6.

Author : Yanghui Ou
  Date : Feb 24, 2019
"""
from collections import deque

from pymtl3.datatypes import Bits16
from pymtl3.dsl.ComponentLevel3 import connect
from pymtl3.dsl.ComponentLevel6 import ComponentLevel6, non_blocking
from pymtl3.dsl.Connectable import CalleeIfcCL, CallerIfcCL
from pymtl3.dsl.ConstraintTypes import M, U

from .sim_utils import simple_sim_pass

#-------------------------------------------------------------------------
# TestSrc
#-------------------------------------------------------------------------

class TestSrc( ComponentLevel6 ):

  def construct( s, msgs ):

    s.send = CallerIfcCL()

    s.msgs = deque( msgs )
    s.head = None
    s.trace_len = len( str( s.msgs[0] ) )

    @s.update
    def send_msg():
      s.head = None
      if s.send.rdy() and s.msgs:
        s.head = s.msgs.popleft()
        s.send( s.head )

  def done( s ):
    return not s.msgs

  def line_trace( s ):
    return "{}".format(
        "" if s.head is None
           else str( s.head ) ).ljust( s.trace_len )

#-------------------------------------------------------------------------
# TestSink
#-------------------------------------------------------------------------

class TestSink( ComponentLevel6 ):

  def construct( s, msgs ):
    s.idx  = 0
    s.msgs = list( msgs )
    s.head = None
    s.trace_len = len( str( s.msgs[0] ) )

  @non_blocking( lambda s: True )
  def recv( s, msg ):

    s.head = msg
    # Sanity check
    if s.idx >= len( s.msgs ):
      raise Exception( "Test Sink received more msgs than expected" )

    if msg != s.msgs[s.idx]:
      raise Exception( """
        Test Sink received WRONG msg!
        Expected : {}
        Received : {}
        """.format( s.msgs[s.idx], msg ) )
    else:
      s.idx += 1

  def done( s ):
    return s.idx >= len( s.msgs )

  def line_trace( s ):
    tmp = s.head
    s.head = None
    return "{}".format(
      "" if tmp is None else str( tmp ) ).ljust( s.trace_len )

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------

class TestHarness( ComponentLevel6 ):

  def construct( s, DUT, src_msgs, sink_msgs ):

    s.src     = TestSrc ( src_msgs  )
    s.sink    = TestSink( sink_msgs )
    s.dut     = DUT()

    # Connections
    connect( s.src.send, s.dut.recv  )
    connect( s.dut.send, s.sink.recv )


  def done( s ):
    return s.src.done() and s.sink.done()

  def line_trace( s ):
    return "{} ->| {} |-> {}".format(
      s.src.line_trace(), s.dut.line_trace(), s.sink.line_trace() )

#-------------------------------------------------------------------------
# run_cl_sim
#-------------------------------------------------------------------------

def run_cl_sim( th, max_cycles=50 ):

  # Create a simulator

  th.elaborate()
  simple_sim_pass( th )

  # Run simluation

  ncycles = 0
  print("{}:{}".format( ncycles, th.line_trace() ))
  while not th.done() and ncycles < max_cycles:
    th.tick()
    ncycles += 1
    print("{}:{}".format( ncycles, th.line_trace() ))

  # Check timeout

  assert ncycles < max_cycles

  th.tick()
  th.tick()
  th.tick()

#-------------------------------------------------------------------------
# SimpleQueue
#-------------------------------------------------------------------------

class SimpleQueue( ComponentLevel6 ):

  def construct( s ):
    s.element = None

    # TODO: Improve this
    s.add_constraints(
      # M( s.deq.method ) < M( s.enq.method ),
      M( s.deq ) < M( s.enq ),
      # M( s.deq.rdy ) < M( s.enq.rdy )
    )

  def empty( s ):
    return s.element is None

  @non_blocking( lambda s : s.empty() )
  def enq( s, ele ):
    s.element = ele

  @non_blocking( lambda s: not s.empty() )
  def deq( s ):
    ret = s.element
    s.element = None
    return ret

  def line_trace( s ):
    return "{}".format(
      "    " if s.element is None else str( s.element ) )

# Test the SimpleQueue as a SW data structure.

def test_queue_sw():

  q = SimpleQueue()
  q.elaborate()

  assert q.enq.rdy()
  assert q.empty()
  assert not q.deq.rdy()

  q.enq( Bits16( 128 ) )
  assert not q.enq.rdy()
  assert not q.empty()
  assert q.deq.rdy()

  assert q.deq() == Bits16( 128 )
  assert q.enq.rdy()
  assert q.empty()
  assert not q.deq.rdy()

#-------------------------------------------------------------------------
# QueueIncr
#-------------------------------------------------------------------------

class QueueIncr( ComponentLevel6 ):

  def construct( s ):
    s.recv  = CalleeIfcCL()
    s.send  = CallerIfcCL()
    s.queue = SimpleQueue()

    connect( s.recv, s.queue.enq )

    s.v = None
    @s.update
    def deq_incr():
      s.v = None
      if s.queue.deq.rdy() and s.send.rdy():
        s.v = s.queue.deq() + 1
        s.send( s.v )

  def line_trace( s ):
    return "{} (+) {}".format( s.queue.line_trace(), s.v )

#-------------------------------------------------------------------------
# QueueIncrChained
#-------------------------------------------------------------------------

class QueueIncrChained( ComponentLevel6 ):

  def construct( s ):

    s.recv = CalleeIfcCL()
    s.send = CallerIfcCL()

    s.q0 = QueueIncr()
    s.q1 = QueueIncr()

    connect( s.recv,    s.q0.recv )
    connect( s.q0.send, s.q1.recv )
    connect( s.q1.send, s.send    )

  def line_trace( s ):
    return "{} -> {}".format( s.q0.line_trace(), s.q1.line_trace() )

#-------------------------------------------------------------------------
# CL tests
#-------------------------------------------------------------------------

def test_queue_incr_cl():
  q    = QueueIncr()
  src_msgs  = [ Bits16( 0 ), Bits16( 1 ), Bits16( 2 ), Bits16( 3 ) ]
  sink_msgs = [ Bits16( 1 ), Bits16( 2 ), Bits16( 3 ), Bits16( 4 ) ]
  th = TestHarness( q, src_msgs, sink_msgs )
  run_cl_sim( th )

def test_chained_queue_incr_cl():
  q    = QueueIncrChained()
  src_msgs  = [ Bits16( 0 ), Bits16( 1 ), Bits16( 2 ), Bits16( 3 ) ]
  sink_msgs = [ Bits16( 2 ), Bits16( 3 ), Bits16( 4 ), Bits16( 5 ) ]
  th = TestHarness( q, src_msgs, sink_msgs )
  run_cl_sim( th )



"""
========================================================================
ComponentLevel6.py
========================================================================
Add method port decorator.

Author : Yanghui Ou, Shunning Jiang
  Date : Jan 24, 2020
"""
from .ComponentLevel5 import ComponentLevel5
from .Connectable import CalleeIfcCL, CalleePort

#-------------------------------------------------------------------------
# non blocking decorator
#-------------------------------------------------------------------------

def non_blocking( rdy, Type=None ):
  def real_decorator( method ):
    method._non_blocking_rdy  = rdy
    method._non_blocking_type = Type
    return method
  return real_decorator

#-------------------------------------------------------------------------
# ComponentLevel6
#-------------------------------------------------------------------------

class ComponentLevel6( ComponentLevel5 ):

  # Override
  def _handle_decorated_methods( s ):

    # The following code handles non-blocking methods
    def bind_method( method ):
      def _bound_method( *args, **kwargs ):
        return method( s, *args, **kwargs )
      return _bound_method

    for x in s.__class__.__dict__:
      method = getattr( s, x )

      # We identify decorated method port here
      if   hasattr( method, "_callee_port" ):
        setattr( s, x, CalleePort( method=method ) )

      # We identify non_blocking methods here
      elif hasattr( method, "_non_blocking_rdy" ):
        rdy  = method._non_blocking_rdy
        Type = method._non_blocking_type
        setattr( s, x, CalleeIfcCL( Type=Type, method=method, rdy=bind_method( rdy ) ) )



#=========================================================================
# TestBehavioralTranslator.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 20, 2019
"""Provide a behavioral translator that fits testing purposes."""

from pymtl3.passes.backends.generic.behavioral import BehavioralTranslator


def mk_TestBehavioralTranslator( _BehavioralTranslator ):
  def make_indent( src, nindent ):
    """Add nindent indention to every line in src."""
    indent = '  '
    for idx, s in enumerate( src ):
      src[ idx ] = nindent * indent + s

  class TestBehavioralTranslator( _BehavioralTranslator ):
    """Testing translator that implements behavioral callback methods."""

    def rtlir_data_type_translation( s, m, dtype ):
      """Translate `dtype` in component `m`.

      This method should be implemented by the structural translator. Since
      we are doing unit testing of the behavioral translator, I added it here
      so everything will work. This method does conflict with the actual
      implementation of `rtlir_data_type_translation` when we start testing
      RTLIRTranslator. My current solution is to explicitly check for the
      test translator's name and if it's a match then call the actual
      implementation.
      """
      if s.__class__.__name__ == 'TestRTLIRTranslator':
        return super(). \
          rtlir_data_type_translation( m, dtype )
      else:
        return str(dtype)

    def rtlir_tr_upblk_decls( s, upblk_decls ):
      decls = ''
      for upblk_decl in sorted(upblk_decls, key=lambda x: x[0]):
        make_indent( upblk_decl, 1 )
        decls += '\n' + '\n'.join( upblk_decl )
      return f'upblk_decls:{decls}\n'

    def rtlir_tr_upblk_decl( s, upblk, src, py_src ):
      return [f'upblk_decl: {upblk.__name__}']

    def rtlir_tr_upblk_py_srcs( s, upblk_py_srcs ):
      py_srcs = ''
      for upblk_py_src in sorted(upblk_py_srcs, key=lambda x: x[0]):
        make_indent( upblk_py_src, 1 )
        py_srcs += '\n' + '\n'.join( upblk_py_src )
      return f'upblk_py_srcs:{py_srcs}\n'

    def rtlir_tr_upblk_py_src( s, upblk, is_lambda, src, lino, filename ):
      return [f'upblk_py_src: {upblk.__name__}']

    def rtlir_tr_upblk_srcs( s, upblk_srcs ):
      srcs = ''
      for upblk_src in sorted(upblk_srcs, key=lambda x: x[0]):
        make_indent( upblk_src, 1 )
        srcs += '\n' + '\n'.join( upblk_src )
      return f'upblk_srcs:{srcs}\n'

    def rtlir_tr_upblk_src( m, upblk, rtlir_upblk ):
      return [f'upblk_src: {rtlir_upblk.name}']

    def rtlir_tr_behavioral_freevars( s, freevars ):
      srcs = ''
      for freevar in sorted(freevars, key=lambda x: x[0]):
        make_indent( freevar, 1 )
        srcs += '\n' + '\n'.join( freevar )
      return f'freevars:{srcs}\n'

    def rtlir_tr_behavioral_freevar( s, id_, rtype, array_type, dtype, obj ):
      return [f'freevar: {id_}']

    def rtlir_tr_behavioral_tmpvars( s, tmpvars ):
      srcs = ''
      for tmpvar in sorted(tmpvars, key=lambda x: x[0]):
        make_indent( tmpvar, 1 )
        srcs += '\n' + '\n'.join( tmpvar )
      return f'tmpvars:{srcs}\n'

    def rtlir_tr_behavioral_tmpvar( s, id_, upblk_id, dtype ):
      return [f'tmpvar: {id_} in {upblk_id} of {dtype}']

    def rtlir_tr_unpacked_array_type( s, array_rtype ):
      """Translate unpacked array type.

      This method should be implemented by the structural translator. Since
      we are doing unit testing of the behavioral translator, I added it here
      so everything will work. To avoid conflicting with the actual implementation
      of `rtlir_tr_unpacked_array_type`, this method returns different values
      depending on the name of the class (i.e. whether we are doing unit testing
      of the behavioral translator or the RTLIR translator).
      """
      if s.__class__.__name__ == 'TestRTLIRTranslator':
        return "" if array_rtype is None else repr(array_rtype)
      else:
        return f'unpacked_array: {array_rtype}'

    def rtlir_tr_vector_dtype( s, dtype ):
      """Translate vector data type.

      This method should be implemented by the structural translator. Since
      we are doing unit testing of the behavioral translator, I added it here
      so everything will work. To avoid conflicting with the actual implementation
      of `rtlir_tr_vector_dtype`, this method returns different values
      depending on the name of the class (i.e. whether we are doing unit testing
      of the behavioral translator or the RTLIR translator).
      """
      if s.__class__.__name__ == 'TestRTLIRTranslator':
        return str( dtype )
      else:
        return f'vector: {dtype}'

  return TestBehavioralTranslator

TestBehavioralTranslator = mk_TestBehavioralTranslator( BehavioralTranslator )



#========================================================================
# CLLineTracePass.py
#========================================================================
# Enable CL line trace.
#
# Author : Yanghui Ou
#   Date : May 21, 2019

from pymtl3.dsl import *
from pymtl3.passes.BasePass import BasePass, PassMetadata


class CLLineTracePass( BasePass ):

  def __init__( self, default_trace_len=8 ):
    self.default_trace_len = default_trace_len

  def __call__( self, top ):
    if not hasattr( top, "_tracing" ):
      top._tracing = PassMetadata()
    top._tracing.clear_cl_trace = self.process_component( top )

  def process_component( self, top ):

    # [wrap_callee_method] wraps the original method in a callee port
    # into a new method that not only calls the origianl method, but
    # also saves the arguments to the method and the return value,
    # which can be used for composing the line trace.
    # The wrapped method also need to update the saved arguments and
    # return value of all the methods this callee port is driving.
    def wrap_callee_method( mport, net ):
      mport.raw_method = mport.method
      def wrapped_method( self, *args, **kwargs ):
        # If it has greenlet i.e. blocking ... we need to make sure
        # we record everything after the method is successfully invoked
        ret = self.raw_method( *args, **kwargs )
        for m in net:
          m.called = True
          m.saved_args = args
          m.saved_kwargs = kwargs
          m.saved_ret = ret
        return ret
      mport.method = lambda *args, **kwargs : wrapped_method( mport, *args, **kwargs )

    # [wrap_caller_method] wraps the original method in a caller port
    # into a new method that calls its driver instead of the actual
    # method, which will trigger the actual driver to update all other
    # method ports connected to this net.
    def wrap_caller_method( mport, driver_method ):
      def wrapped_method( self, *args, **kwargs ):
        return driver_method( *args, **kwargs )
      mport.method = lambda *args, **kwargs : wrapped_method( mport, *args, **kwargs )

    # Collect all method ports and add some stamps
    all_callees = set()
    all_method_ports = top.get_all_object_filter(
      lambda s: isinstance( s, MethodPort )
    )
    for mport in all_method_ports:
      mport.called = False
      mport.saved_args = None
      mport.saved_kwargs = None
      mport.saved_ret = None
      if isinstance( mport, CalleePort ):
        all_callees.add( mport )

    # Collect all method nets and wrap the actual driving method
    all_drivers = set()
    all_method_nets = top.get_all_method_nets()
    for driver, net in all_method_nets:
      if driver is not None:
        wrap_callee_method( driver, net )
        all_drivers.add( driver )
      for member in net:
        if isinstance( member, CallerPort ):
          assert member is not driver
          wrap_caller_method( member, driver )

    # Handle other callee that is not driving anything
    for mport in ( all_callees - all_drivers ):
      wrap_callee_method( mport, set() )

    # [mk_new_str] replaces [_str_hook] in a non-blocking interface with
    # a new to-string function that uses the metadata to compose line
    # trace.
    # When the rdy is called and returns true, and the method gets called,
    # the line trace just prints out the actual message. Otherwise, it '
    # prints out some special characters under different circumstances:
    # - 'x' rdy not called but method called
    # - '.' rdy not called, method not called
    # - "#" rdy called and is false, method not called
    # - "X" rdy called and is false, method still called
    # - " " rdy called and is true, method not called
    # For example, a cycle-level single element normal queue would have
    # the following line trace:
    #      enq     deq
    #  1:( 0000 () #    ) - enq(0000) called, deq is not ready
    #  2:( #    () 0000 ) - enq is not ready, deq() gets called
    #  3:( 0001 () #    ) - enq(0001) called, deq is not ready again

    def mk_new_str_non_blocking( ifc ):
      def new_str():
        # If rdy is called
        if ifc.rdy.called:
          # If rdy is called and returns true
          if ifc.rdy.saved_ret:
            # If rdy and method called - return actual message
            if ifc.method.called:
              args_strs = [ str( arg ) for arg in ifc.method.saved_args ] + \
                          [ str( arg ) for _, arg in ifc.method.saved_kwargs.items() ]

              ret_str = "" if ifc.method.saved_ret is None else str( ifc.method.saved_ret )

              trace = ""
              if args_strs:
                trace += f"({','.join(args_strs)})"
              if ret_str:
                trace += f"={ret_str}"

              ifc.trace_len = len(trace)
              return trace

            # If rdy and method not called
            else:
              return " ".ljust( ifc.trace_len )

          # If rdy is called and returns false
          elif ifc.method.called:
            return "X".ljust( ifc.trace_len )
          else:
            return "#".ljust( ifc.trace_len )

        # If rdy is not called
        elif ifc.method.called:
          return "x".ljust( ifc.trace_len )

        else:
          return ".".ljust( ifc.trace_len )
      return new_str

    # Collecting all non blocking interfaces and replace the str hook
    for ifc in top.get_all_object_filter( lambda s: isinstance( s, NonBlockingIfc ) ):
      if ifc.method.Type is not None:
        ifc.trace_len = len( str( ifc.method.Type() ) )
      else:
        ifc.trace_len = self.default_trace_len
      ifc._str_hook = mk_new_str_non_blocking( ifc )

    # [mk_new_str] replaces [_str_hook] in a blocking interface with
    # a new to-string function that uses the metadata to compose line
    # trace. The case for blocking interfaces is simpler than
    # non-blocking interfaces as we only have two possibilities
    # called/not called.
    # - " " method not called
    # - msg method called
    def mk_new_str_blocking( ifc ):
      def new_str():
        # If method called - return actual message
        if ifc.method.called:
          args_strs = [ str( arg ) for arg in ifc.method.saved_args ] + \
                      [ str( arg ) for _, arg in ifc.method.saved_kwargs.items() ]

          ret_str = "" if ifc.method.saved_ret is None else str( ifc.method.saved_ret )

          trace = ""
          if args_strs:
            trace += f"({','.join(args_strs)})"
          if ret_str:
            trace += f"={ret_str}"

          ifc.trace_len = len(trace)
          return trace

        # If method not called
        else:
          return " ".ljust( ifc.trace_len )
      return new_str

    # Collecting all blocking interfaces and replace the str hook
    for ifc in top.get_all_object_filter( lambda s: isinstance( s, BlockingIfc ) ):
      if ifc.method.Type is not None:
        ifc.trace_len = len( str( ifc.method.Type() ) )
      else:
        ifc.trace_len = self.default_trace_len
      ifc._str_hook = mk_new_str_blocking( ifc )

    # An update block that resets all method ports to not called
    def reset_method_ports():
      for mport in all_method_ports:
        mport.called = False
        mport.saved_args = None
        mport.saved_kwargs = None
        mport.saved_ret = None

    return reset_method_ports



"""
========================================================================
Encoder_test
========================================================================
Test for RTL priority encoder.

Author : Yanghui Ou
  Date : Apr 5, 2019
"""
from pymtl3 import *
from pymtl3.stdlib.test import TestVectorSimulator

from .Encoder import Encoder


def run_test( model, test_vectors ):

  def tv_in( model, test_vector ):
    model.in_ = test_vector[0]

  def tv_out( model, test_vector ):
    assert model.out == test_vector[1]

  sim = TestVectorSimulator( model, test_vectors, tv_in, tv_out )
  sim.run_test()

def test_encoder_5_directed():

  model = Encoder( 5, 3 )
  run_test( model, [
    # in                out
    [ Bits5( 0b00000 ), 0  ],
    [ Bits5( 0b00001 ), 0  ],
    [ Bits5( 0b00010 ), 1  ],
    [ Bits5( 0b00100 ), 2  ],
    [ Bits5( 0b01100 ), 3  ],
    [ Bits5( 0b10000 ), 4  ]
  ] )



#=========================================================================
# SVBehavioralTranslatorL0.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 18, 2019
"""Provide the level 0 SystemVerilog translator implementation."""

from pymtl3.passes.backends.generic.behavioral.BehavioralTranslatorL0 import (
    BehavioralTranslatorL0,
)


class SVBehavioralTranslatorL0( BehavioralTranslatorL0 ):
  pass
